from functools import wraps
from django.db import transaction, DEFAULT_DB_ALIAS

class Transaction(object):
    sid = None

    def __init__(self, using=None):
        self.using = using

    def __enter__(self):
        if transaction.is_managed():
            self.sid = transaction.savepoint()
            pass
        else:
            transaction.enter_transaction_management(using=self.using)

    def __exit__(self, exc_type, exc_value, traceback):
        if self.sid is not None:
            if exc_value is not None:
                transaction.savepoint_rollback(self.sid, using=self.using)
        else:
            try:
                if exc_value is not None:
                    if transaction.is_dirty(using=self.using):
                        transaction.rollback(using=self.using)
                else:
                    if transaction.is_dirty(using=self.using):
                        try:
                            transaction.commit(using=self.using)
                        except:
                            transaction.rollback(using=self.using)
                            raise
            finally:
                transaction.leave_transaction_management(using=self.using)

    def __call__(self, func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            with self.__class__(using=self.using):
                return func(*args, **kwargs)
        return wrapper


def commit_on_success(using=None):
    """
    This decorator activates commit on response. This way, if the view function
    runs successfully, a commit is made; if the viewfunc produces an exception,
    a rollback is made. This is one of the most common ways to do transaction
    control in Web apps.
    """
    if using is None:
        using = DEFAULT_DB_ALIAS
    if callable(using):
        return Transaction(DEFAULT_DB_ALIAS)(using)
    return Transaction(using)





from django.contrib.contenttypes.models import ContentType
from django.db.models.query import QuerySet
from django.core.cache import cache
from model_utils.managers import PassThroughManager

import models


JOURNAL_METADATA_CACHE_TIMEOUT = 3600


class CachedQuerySet(QuerySet):
    def get_cached(self, **kwargs):
        key = tuple(sorted(kwargs.iteritems()))
        hash1 = key.__hash__()
        hash2 = ('x', key).__hash__()
        key = 'dj-cache-%s-%s-%s' % (self.model.__name__, hash1, hash2)
        instance = cache.get(key)
        if instance is not None:
            return instance
        instance, created = self.get_or_create(**kwargs)
        if JOURNAL_METADATA_CACHE_TIMEOUT:
            cache.set(key, instance, JOURNAL_METADATA_CACHE_TIMEOUT)
        return instance


CachedManager = PassThroughManager.for_queryset_class(CachedQuerySet)


class TagManager(CachedManager):
    def get_by_natural_key(self, name):
        return self.get(name=name)


class TemplateManager(CachedManager):
    def get_by_natural_key(self, content):
        return self.get(content=content)


class JournalQuerySet(QuerySet):
    def for_object(self, obj, tag=None):
        '''Return Journal records linked to this object.'''
        content_type = ContentType.objects.get_for_model(obj)
        if tag is None:
            return self.filter(objectdata__content_type=content_type,
                    objectdata__object_id=obj.pk)
        else:
            return self.filter(
                    objectdata__tag__name=tag,
                    objectdata__content_type=content_type,
                    objectdata__object_id=obj.pk)

    def for_objects(self, objects):
        '''Return journal records linked to any of this objects.

           All objects must have the same model.
        '''
        if not objects:
            return self.none()
        content_types = [ ContentType.objects.get_for_model(obj)
                for obj in objects ]
        if len(set(content_types)) != 1:
            raise ValueError('objects must have of the same content type')
        pks = [ obj.pk for obj in objects ]
        return self.filter(
                objectdata__content_type=content_types[0],
                objectdata__object_id__in=pks)

    def for_tag(self, tag):
        '''Returns Journal records linked to this tag by their own tag or
           the tag on their data records.
        '''
        if not isinstance(tag, models.Tag):
            try:
                tag = models.Tag.objects.get_cached(name=tag)
            except models.Tag.DoesNotExist:
                return self.none()
        # always remember: multiple join (OR in WHERE) produces duplicate
        # lines ! Use .distinct() for safety.
        return self.filter(models.Q(tag=tag)|
                models.Q(objectdata__tag=tag)|
                models.Q(stringdata__tag=tag)) \
                .distinct()


class JournalManager(PassThroughManager.for_queryset_class(JournalQuerySet)):
    def get_query_set(self):
        return super(JournalManager, self).get_query_set() \
               .prefetch_related('objectdata_set__content_type',
                       'stringdata_set', 'objectdata_set__tag',
                       'stringdata_set__tag', 'objectdata_set__content_object',
                       'tag', 'template') \
               .select_related('tag', 'template')



import csv

from django.utils.translation import ugettext_lazy as _
from django.http import HttpResponse

def export_as_csv(modeladmin, request, queryset):
    """
    CSV export for journal
    """
    response = HttpResponse(mimetype='text/csv')
    response['Content-Disposition'] = 'attachment; filename=journal.csv'
    writer = csv.writer(response)

    writer.writerow(['time', 'tag', 'message'])
    for journal in queryset:
        row = [
                journal.time.isoformat(' '),
                journal.tag.name.encode('utf-8'),
                unicode(journal).encode('utf-8'),
              ]
        writer.writerow(row)
    return response
export_as_csv.short_description = _(u"Export CSV file")



from django.test import TestCase
from django.contrib.auth.models import User, Group
from django.db import transaction

from . import record
import models

class JournalTestCase(TestCase):
    def setUp(self):
        models.JOURNAL_METADATA_CACHE_TIMEOUT = 0
        self.users = []
        self.groups = []
        with transaction.commit_on_success():
            for i in range(20):
                self.users.append(
                        User.objects.create(username='user%s' % i))
            for i in range(20):
                self.groups.append(
                        Group.objects.create(name='group%s' % i))
            for i in range(20):
                record('login', '{user} logged in', user=self.users[i])
            for i in range(20):
                record('group-changed', '{user1} gave group {group} to {user2}',
                        user1=self.users[i], group=self.groups[i], 
                        user2=self.users[(i+1) % 20])
            for i in range(20):
                record('logout', '{user} logged out', user=self.users[i])

    def test_login(self):
        for i, event in zip(range(20), models.Journal.objects.for_tag('login').order_by('id')):
            self.assertEqual(unicode(event), 'user{0} logged in'.format(i))

    def test_groups(self):
        for i, event in zip(range(40), models.Journal.objects.for_tag('group-changed').order_by('id')):
            self.assertEqual(unicode(event),
                    'user{0} gave group group{0} to user{1}'.format(i, (i+1)%20))

    def test_logout(self):
        for i, event in zip(range(20), models.Journal.objects.for_tag('logout').order_by('id')):
            self.assertEqual(unicode(event), 'user{0} logged out'.format(i))




import django_journal

class JournalMiddleware(object):
    '''Add record and error_record methods to the request object to log
       current user and current REMOTE_ADRESS.

       It must be setup after the auth middleware.
    '''

    def process_request(self, request):
        user = getattr(request, 'user', None)
        ip = request.META.get('REMOTE_ADDR', None)
        def record(tag, template, using=None, **kwargs):
            if 'user' not in kwargs:
                kwargs['user'] = user
            if 'ip' not in kwargs:
                kwargs['ip'] = ip
            django_journal.record(tag, template, using=using,**kwargs)
        def error_record(tag, template, using=None, **kwargs):
            if 'user' not in kwargs:
                kwargs['user'] = user
            if 'ip' not in kwargs:
                kwargs['ip'] = ip
            django_journal.error_record(tag, template, using=using, **kwargs)
        request.record = record
        request.error_record = error_record
        return None

    def process_response(self, request, response):
        if hasattr(request, 'record'):
            del request.record
        if hasattr(request, 'error_record'):
            del request.error_record
        return response



import string

from django.db import models
from django.contrib.contenttypes import generic
from django.utils.translation import ugettext_lazy as _

import managers


class Tag(models.Model):
    '''Tag allows typing event and data linked to events.

       name:
           the string identifier of the tag
    '''
    objects = managers.TagManager()
    name = models.CharField(verbose_name=_('name'), max_length=32, unique=True,
            db_index=True)

    def __unicode__(self):
        return self.name

    def natural_key(self):
        return (self.name,)

    class Meta:
        ordering = ('name',)
        verbose_name = _('tag')


class Template(models.Model):
    '''Template for formatting an event.

       ex.: Template(
                content='{user1} gave group {group} to {user2}')
    '''
    objects = managers.TemplateManager()
    content = models.TextField(verbose_name=_('content'), unique=True,
            db_index=True)

    def __unicode__(self):
        return self.content

    def natural_key(self):
        return (self.content,)

    class Meta:
        ordering = ('content',)


class Journal(models.Model):
    '''One line of the journal.

       Each recorded event in the journal is a Journal instance.

       time - the time at which the event was recorded
       tag - the tag giving the type of event
       template - a format string to present the event
       message - a simple string representation of the event, computed using
       the template and associated datas.
    '''
    objects = managers.JournalManager()

    time = models.DateTimeField(verbose_name=_('time'), auto_now_add=True,
            db_index=True)
    tag = models.ForeignKey(Tag, verbose_name=_('tag'),
            on_delete=models.PROTECT)
    template = models.ForeignKey(Template, verbose_name=_('template'),
            on_delete=models.PROTECT)
    message = models.CharField(verbose_name=_('message'), max_length=128,
            db_index=True)

    class Meta:
        ordering = ('-id',)
        verbose_name = _('journal entry')
        verbose_name_plural = _('journal entries')

    def message_context(self):
        ctx = {}
        for data in self.objectdata_set.all():
            if data.content_object is not None:
                ctx[data.tag.name] = data.content_object
            else:
                ctx[data.tag.name] = u'<deleted {content_type} {object_id}>'.format(
                        content_type=data.content_type, object_id=data.object_id)
        for data in self.stringdata_set.all():
            ctx[data.tag.name] = data.content
        for text, field, format_spec, conversion in string.Formatter().parse(self.template.content):
            if not field:
                continue
            splitted = field.split('.')
            if splitted[0] not in ctx:
                ctx[splitted[0]] = None
        return ctx

    def add_object_tag(self, tag_name, obj):
        ObjectData(journal=self,
                tag=Tag.objects.get_cached(name=tag_name),
                content_object=obj).save()

    def __unicode__(self):
        ctx = self.message_context()
        return self.template.content.format(**ctx)

    def __repr__(self):
        return '<Journal pk:{0} tag:{1} message:{2}>'.format(
                self.pk, unicode(self.tag).encode('utf-8'),
                unicode(self.message).encode('utf-8'))


class StringData(models.Model):
    '''String data associated to a recorded event.

       journal:
           the recorded event
       tag:
           the identifier for this data
       content:
           the string value of the data
    '''
    journal = models.ForeignKey(Journal, verbose_name=_('journal entry'))
    tag = models.ForeignKey(Tag, verbose_name=_('tag'))
    content = models.TextField(verbose_name=_('content'))

    class Meta:
        unique_together = (('journal', 'tag'),)
        verbose_name = _('linked text string')


class ObjectData(models.Model):
    '''Object data associated with a recorded event.

       journal:
           the recorded event
       tag:
           the identifier for this data
       content_object:
           the object value of the data
    '''
    journal = models.ForeignKey(Journal, verbose_name=_('journal entry'))
    tag = models.ForeignKey(Tag, verbose_name=_('tag'))
    content_type = models.ForeignKey('contenttypes.ContentType',
            verbose_name=_('content type'))
    object_id = models.PositiveIntegerField(db_index=True,
            verbose_name=_('object id'))
    content_object = generic.GenericForeignKey('content_type',
            'object_id')

    class Meta:
        unique_together = (('journal', 'tag'),)
        verbose_name = _('linked object')

    def __unicode__(self):
        return u'{0}:{1}:{2}'.format(self.journal.id, self.tag, self.content_object)



import logging

from exceptions import JournalException
from models import (Journal, Tag, Template)

import django.db.models
from django.conf import settings

from decorator import commit_on_success

__all__ = ('record', 'error_record', 'Journal')
__version__ = '1.19.0'

def unicode_truncate(s, length, encoding='utf-8'):
    '''Truncate an unicode string so that its UTF-8 encoding is less than
       length.'''
    encoded = s.encode(encoding)[:length]
    return encoded.decode(encoding, 'ignore')

@commit_on_success
def record(tag, template, using=None, **kwargs):
    '''Record an event in the journal. The modification is done inside the
       current transaction.

       tag:
           a string identifier giving the type of the event
       tpl:
           a format string to describe the event
       kwargs:
           a mapping of object or data to interpolate in the format string
    '''
    template = unicode(template)
    tag = Tag.objects.using(using).get_cached(name=tag)
    template = Template.objects.using(using).get_cached(content=template)
    try:
        message = template.content.format(**kwargs)
    except (KeyError, IndexError), e:
        raise JournalException(
                'Missing variable for the template message', template, e)
    try:
        logger = logging.getLogger('django.journal.%s' % tag)
        if tag.name == 'error':
            logger.error(message)
        else:
            logger.info(message)
    except:
        try:
            logging.getLogger('django.journal').exception('Unable to log msg')
        except:
            pass # we tried, really, we tried
    journal = Journal.objects.using(using).create(tag=tag, template=template,
            message=unicode_truncate(message, 128))
    for tag, value in kwargs.iteritems():
        if value is None:
            continue
        tag = Tag.objects.using(using).get_cached(name=tag)
        if isinstance(value, django.db.models.Model):
            journal.objectdata_set.create(tag=tag, content_object=value)
        else:
            journal.stringdata_set.create(tag=tag, content=unicode(value))
    return journal

def error_record(tag, tpl, **kwargs):
    '''Records error events.

       You must use this function when logging error events. It uses another
       database alias than the default one to be immune to transaction rollback
       when logging in the middle of a transaction which is going to
       rollback.
    '''
    return record(tag, tpl,
            using=getattr(settings, 'JOURNAL_DB_FOR_ERROR_ALIAS', 'default'),
            **kwargs)



#!/usr/bin/python
import os

from setuptools import setup, find_packages
from setuptools.command.install_lib import install_lib as _install_lib
from distutils.command.build import build as _build
from distutils.command.sdist import sdist  as _sdist
from distutils.cmd import Command

class compile_translations(Command):
    description = 'compile message catalogs to MO files via django compilemessages'
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        import os
        import sys
        from django.core.management.commands.compilemessages import \
            compile_messages
        curdir = os.getcwd()
        os.chdir(os.path.realpath('django_journal'))
        compile_messages(stderr=sys.stderr)
        os.chdir(curdir)

class build(_build):
    sub_commands = [('compile_translations', None)] + _build.sub_commands

class sdist(_sdist):
    sub_commands = [('compile_translations', None)] + _sdist.sub_commands

class install_lib(_install_lib):
    def run(self):
        self.run_command('compile_translations')
        _install_lib.run(self)

def get_version():
    import glob
    import re
    import os

    version = None
    for d in glob.glob('*'):
        if not os.path.isdir(d):
            continue
        module_file = os.path.join(d, '__init__.py')
        if not os.path.exists(module_file):
            continue
        for v in re.findall("""__version__ *= *['"](.*)['"]""",
                open(module_file).read()):
            assert version is None
            version = v
        if version:
            break
    assert version is not None
    if os.path.exists('.git'):
        import subprocess
        p = subprocess.Popen(['git','describe','--dirty', '--match=v*'],
                stdout=subprocess.PIPE)
        result = p.communicate()[0]
        assert p.returncode == 0, 'git returned non-zero'
        new_version = result.split()[0][1:]
        major_minor_release = new_version.split('-')[0]
        assert  version == major_minor_release, \
                '__version__ (%s) must match the last git annotated tag (%s)' % (version, major_minor_release)
        version = new_version.replace('-', '.')
    return version

setup(name='django-journal',
        version=get_version(),
        license='AGPLv3',
        description='Keep a structured -- i.e. not just log strings -- journal'
                    ' of events in your applications',
        url='http://dev.entrouvert.org/projects/django-journal/',
        download_url='http://repos.entrouvert.org/django-journal.git/',
        author="Entr'ouvert",
        author_email="info@entrouvert.com",
        packages=find_packages(os.path.dirname(__file__) or '.'),
        include_package_data=True,
        cmdclass={'build': build, 'install_lib': install_lib,
            'compile_translations': compile_translations,
            'sdist': sdist},
        install_requires=[
            'django >= 1.4.2',
            'django-model-utils<1.4',
        ],
        setup_requires=[
            'django >= 1.4.2',
        ],
)
